@@ -3,10 +3,10 @@ package main
 import (
 	"crypto/sha512"
 	"encoding/binary"
-	"sync"
+	"fmt"
 	"time"
 
-	lru "github.com/hashicorp/golang-lru"
+	"github.com/jedisct1/go-sieve-cache/pkg/sievecache"
 	"github.com/miekg/dns"
 )
 
@@ -18,8 +18,7 @@ type CachedResponse struct {
 }
 
 type CachedResponses struct {
-	sync.RWMutex
-	cache *lru.ARCCache
+	cache *sievecache.ShardedSieveCache[[32]byte, CachedResponse]
 }
 
 var cachedResponses CachedResponses
@@ -45,8 +44,7 @@ func computeCacheKey(pluginsState *PluginsState, msg *dns.Msg) [32]byte {
 
 // ---
 
-type PluginCache struct {
-}
+type PluginCache struct{}
 
 func (plugin *PluginCache) Name() string {
 	return "cache"
@@ -71,20 +69,15 @@ func (plugin *PluginCache) Reload() error {
 func (plugin *PluginCache) Eval(pluginsState *PluginsState, msg *dns.Msg) error {
 	cacheKey := computeCacheKey(pluginsState, msg)
 
-	cachedResponses.RLock()
 	if cachedResponses.cache == nil {
-		cachedResponses.RUnlock()
 		return nil
 	}
-	cachedAny, ok := cachedResponses.cache.Get(cacheKey)
+	cached, ok := cachedResponses.cache.Get(cacheKey)
 	if !ok {
-		cachedResponses.RUnlock()
 		return nil
 	}
-	cached := cachedAny.(CachedResponse)
 	expiration := cached.expiration
 	synth := cached.msg.Copy()
-	cachedResponses.RUnlock()
 
 	synth.Id = msg.Id
 	synth.Response = true
@@ -108,8 +101,7 @@ func (plugin *PluginCache) Eval(pluginsState *PluginsState, msg *dns.Msg) error
 
 // ---
 
-type PluginCacheResponse struct {
-}
+type PluginCacheResponse struct{}
 
 func (plugin *PluginCacheResponse) Name() string {
 	return "cache_response"
@@ -150,17 +142,14 @@ func (plugin *PluginCacheResponse) Eval(pluginsState *PluginsState, msg *dns.Msg
 		expiration: time.Now().Add(ttl),
 		msg:        *msg,
 	}
-	cachedResponses.Lock()
 	if cachedResponses.cache == nil {
-		var err error
-		cachedResponses.cache, err = lru.NewARC(pluginsState.cacheSize)
+		cache, err := sievecache.NewSharded[[32]byte, CachedResponse](pluginsState.cacheSize)
 		if err != nil {
-			cachedResponses.Unlock()
-			return err
+			return fmt.Errorf("failed to initialize the cache: %w", err)
 		}
+		cachedResponses.cache = cache
 	}
-	cachedResponses.cache.Add(cacheKey, cachedResponse)
-	cachedResponses.Unlock()
+	cachedResponses.cache.Insert(cacheKey, cachedResponse)
 	updateTTL(msg, cachedResponse.expiration)
 
 	return nil
