@@ -4,7 +4,6 @@ import (
 	"context"
 	crypto_rand "crypto/rand"
 	"encoding/binary"
-	"math/rand"
 	"net"
 	"os"
 	"runtime"
@@ -15,7 +14,6 @@ import (
 	"github.com/jedisct1/dlog"
 	clocksmith "github.com/jedisct1/go-clocksmith"
 	stamps "github.com/jedisct1/go-dnsstamps"
-	"github.com/miekg/dns"
 	"golang.org/x/crypto/curve25519"
 )
 
@@ -31,12 +29,15 @@ type Proxy struct {
 	queryLogIgnoredQtypes         []string
 	localDoHListeners             []*net.TCPListener
 	queryMeta                     []string
+	enableHotReload               bool
 	udpListeners                  []*net.UDPConn
 	sources                       []*Source
 	tcpListeners                  []*net.TCPListener
 	registeredRelays              []RegisteredServer
 	listenAddresses               []string
 	localDoHListenAddresses       []string
+	monitoringUI                  MonitoringUIConfig
+	monitoringInstance            *MonitoringUI
 	xTransport                    *XTransport
 	allWeeklyRanges               *map[string]WeeklyRanges
 	routes                        *map[string][]string
@@ -74,6 +75,7 @@ type Proxy struct {
 	certRefreshDelayAfterFailure  time.Duration
 	timeout                       time.Duration
 	certRefreshDelay              time.Duration
+	certRefreshConcurrency        int
 	cacheSize                     int
 	logMaxBackups                 int
 	logMaxAge                     int
@@ -117,11 +119,18 @@ func (proxy *Proxy) registerLocalDoHListener(listener *net.TCPListener) {
 }
 
 func (proxy *Proxy) addDNSListener(listenAddrStr string) {
-	listenUDPAddr, err := net.ResolveUDPAddr("udp", listenAddrStr)
+	udp := "udp"
+	tcp := "tcp"
+	isIPv4 := isDigit(listenAddrStr[0])
+	if isIPv4 {
+		udp = "udp4"
+		tcp = "tcp4"
+	}
+	listenUDPAddr, err := net.ResolveUDPAddr(udp, listenAddrStr)
 	if err != nil {
 		dlog.Fatal(err)
 	}
-	listenTCPAddr, err := net.ResolveTCPAddr("tcp", listenAddrStr)
+	listenTCPAddr, err := net.ResolveTCPAddr(tcp, listenAddrStr)
 	if err != nil {
 		dlog.Fatal(err)
 	}
@@ -140,11 +149,11 @@ func (proxy *Proxy) addDNSListener(listenAddrStr string) {
 	// if 'userName' is set and we are the parent process
 	if !proxy.child {
 		// parent
-		listenerUDP, err := net.ListenUDP("udp", listenUDPAddr)
+		listenerUDP, err := net.ListenUDP(udp, listenUDPAddr)
 		if err != nil {
 			dlog.Fatal(err)
 		}
-		listenerTCP, err := net.ListenTCP("tcp", listenTCPAddr)
+		listenerTCP, err := net.ListenTCP(tcp, listenTCPAddr)
 		if err != nil {
 			dlog.Fatal(err)
 		}
@@ -185,7 +194,12 @@ func (proxy *Proxy) addDNSListener(listenAddrStr string) {
 }
 
 func (proxy *Proxy) addLocalDoHListener(listenAddrStr string) {
-	listenTCPAddr, err := net.ResolveTCPAddr("tcp", listenAddrStr)
+	network := "tcp"
+	isIPv4 := isDigit(listenAddrStr[0])
+	if isIPv4 {
+		network = "tcp4"
+	}
+	listenTCPAddr, err := net.ResolveTCPAddr(network, listenAddrStr)
 	if err != nil {
 		dlog.Fatal(err)
 	}
@@ -201,7 +215,7 @@ func (proxy *Proxy) addLocalDoHListener(listenAddrStr string) {
 	// if 'userName' is set and we are the parent process
 	if !proxy.child {
 		// parent
-		listenerTCP, err := net.ListenTCP("tcp", listenTCPAddr)
+		listenerTCP, err := net.ListenTCP(network, listenTCPAddr)
 		if err != nil {
 			dlog.Fatal(err)
 		}
@@ -232,6 +246,23 @@ func (proxy *Proxy) StartProxy() {
 		dlog.Fatal(err)
 	}
 	curve25519.ScalarBaseMult(&proxy.proxyPublicKey, &proxy.proxySecretKey)
+
+	// Initialize and start the monitoring UI if enabled
+	if proxy.monitoringUI.Enabled {
+		dlog.Noticef("Initializing monitoring UI")
+		proxy.monitoringInstance = NewMonitoringUI(proxy)
+		if proxy.monitoringInstance == nil {
+			dlog.Errorf("Failed to create monitoring UI instance")
+		} else {
+			dlog.Noticef("Starting monitoring UI")
+			if err := proxy.monitoringInstance.Start(); err != nil {
+				dlog.Errorf("Failed to start monitoring UI: %v", err)
+			} else {
+				dlog.Noticef("Monitoring UI started successfully")
+			}
+		}
+	}
+
 	proxy.startAcceptingClients()
 	if !proxy.child {
 		// Notify the service manager that dnscrypt-proxy is ready. dnscrypt-proxy manages itself in case
@@ -242,6 +273,8 @@ func (proxy *Proxy) StartProxy() {
 			dlog.Fatal(err)
 		}
 	}
+	proxy.xTransport.internalResolverReady = false
+	proxy.xTransport.internalResolvers = proxy.listenAddresses
 	liveServers, err := proxy.serversInfo.refresh(proxy)
 	if liveServers > 0 {
 		proxy.certIgnoreTimestamp = false
@@ -439,7 +472,13 @@ func (proxy *Proxy) udpListenerFromAddr(listenAddr *net.UDPAddr) error {
 	if err != nil {
 		return err
 	}
-	clientPc, err := listenConfig.ListenPacket(context.Background(), "udp", listenAddr.String())
+	listenAddrStr := listenAddr.String()
+	network := "udp"
+	isIPv4 := isDigit(listenAddrStr[0])
+	if isIPv4 {
+		network = "udp4"
+	}
+	clientPc, err := listenConfig.ListenPacket(context.Background(), network, listenAddrStr)
 	if err != nil {
 		return err
 	}
@@ -453,7 +492,13 @@ func (proxy *Proxy) tcpListenerFromAddr(listenAddr *net.TCPAddr) error {
 	if err != nil {
 		return err
 	}
-	acceptPc, err := listenConfig.Listen(context.Background(), "tcp", listenAddr.String())
+	listenAddrStr := listenAddr.String()
+	network := "tcp"
+	isIPv4 := isDigit(listenAddrStr[0])
+	if isIPv4 {
+		network = "tcp4"
+	}
+	acceptPc, err := listenConfig.Listen(context.Background(), network, listenAddrStr)
 	if err != nil {
 		return err
 	}
@@ -467,7 +512,13 @@ func (proxy *Proxy) localDoHListenerFromAddr(listenAddr *net.TCPAddr) error {
 	if err != nil {
 		return err
 	}
-	acceptPc, err := listenConfig.Listen(context.Background(), "tcp", listenAddr.String())
+	listenAddrStr := listenAddr.String()
+	network := "tcp"
+	isIPv4 := isDigit(listenAddrStr[0])
+	if isIPv4 {
+		network = "tcp4"
+	}
+	acceptPc, err := listenConfig.Listen(context.Background(), network, listenAddrStr)
 	if err != nil {
 		return err
 	}
@@ -515,7 +566,7 @@ func (proxy *Proxy) exchangeWithUDPServer(
 	var pc net.Conn
 	proxyDialer := proxy.xTransport.proxyDialer
 	if proxyDialer == nil {
-		pc, err = net.DialUDP("udp", nil, upstreamAddr)
+		pc, err = net.DialTimeout("udp", upstreamAddr.String(), serverInfo.Timeout)
 	} else {
 		pc, err = (*proxyDialer).Dial("udp", upstreamAddr.String())
 	}
@@ -558,7 +609,7 @@ func (proxy *Proxy) exchangeWithTCPServer(
 	var pc net.Conn
 	proxyDialer := proxy.xTransport.proxyDialer
 	if proxyDialer == nil {
-		pc, err = net.DialTCP("tcp", nil, upstreamAddr)
+		pc, err = net.DialTimeout("tcp", upstreamAddr.String(), serverInfo.Timeout)
 	} else {
 		pc, err = (*proxyDialer).Dial("tcp", upstreamAddr.String())
 	}
@@ -601,10 +652,16 @@ func (proxy *Proxy) clientsCountInc() bool {
 
 func (proxy *Proxy) clientsCountDec() {
 	for {
-		if count := atomic.LoadUint32(&proxy.clientsCount); count == 0 ||
-			atomic.CompareAndSwapUint32(&proxy.clientsCount, count, count-1) {
+		count := atomic.LoadUint32(&proxy.clientsCount)
+		if count == 0 {
+			// Already at zero, nothing to do
 			break
 		}
+		if atomic.CompareAndSwapUint32(&proxy.clientsCount, count, count-1) {
+			dlog.Debugf("clients count: %d", count-1)
+			break
+		}
+		// CAS failed, retry with updated count
 	}
 }
 
@@ -617,11 +674,19 @@ func (proxy *Proxy) processIncomingQuery(
 	start time.Time,
 	onlyCached bool,
 ) []byte {
-	var response []byte = nil
-	if len(query) < MinDNSPacketSize {
+	// Initialize metrics for this query
+	dlog.Debugf("Processing incoming query from %s", (*clientAddr).String())
+
+	// Validate the query
+	var response []byte
+	if !validateQuery(query) {
 		return response
 	}
+
+	// Initialize plugin state
 	pluginsState := NewPluginsState(proxy, clientProto, clientAddr, serverProto, start)
+
+	// Get server info and initialize parameters
 	serverName := "-"
 	needsEDNS0Padding := false
 	serverInfo := proxy.serversInfo.getOne()
@@ -629,201 +694,59 @@ func (proxy *Proxy) processIncomingQuery(
 		serverName = serverInfo.Name
 		needsEDNS0Padding = (serverInfo.Proto == stamps.StampProtoTypeDoH || serverInfo.Proto == stamps.StampProtoTypeTLS)
 	}
+
+	// Apply query plugins
 	query, _ = pluginsState.ApplyQueryPlugins(&proxy.pluginsGlobals, query, needsEDNS0Padding)
-	if len(query) < MinDNSPacketSize || len(query) > MaxDNSPacketSize {
+	if !validateQuery(query) {
 		return response
 	}
+
+	// Handle query plugin actions
 	if pluginsState.action == PluginsActionDrop {
 		pluginsState.returnCode = PluginsReturnCodeDrop
 		pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
 		return response
 	}
+
+	// Handle synthesized responses from plugins
 	var err error
 	if pluginsState.synthResponse != nil {
-		response, err = pluginsState.synthResponse.PackBuffer(response)
+		response, err = handleSynthesizedResponse(&pluginsState, pluginsState.synthResponse)
 		if err != nil {
-			pluginsState.returnCode = PluginsReturnCodeParseError
-			pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
 			return response
 		}
 	}
+
+	// Return early if only cached results are requested
 	if onlyCached {
 		if len(response) == 0 {
 			return response
 		}
 		serverInfo = nil
 	}
+
+	// Process query with a DNS server if there's no cached response
 	if len(response) == 0 && serverInfo != nil {
-		var ttl *uint32
 		pluginsState.serverName = serverName
-		if serverInfo.Proto == stamps.StampProtoTypeDNSCrypt {
-			sharedKey, encryptedQuery, clientNonce, err := proxy.Encrypt(serverInfo, query, serverProto)
-			if err != nil && serverProto == "udp" {
-				dlog.Debug("Unable to pad for UDP, re-encrypting query for TCP")
-				serverProto = "tcp"
-				sharedKey, encryptedQuery, clientNonce, err = proxy.Encrypt(serverInfo, query, serverProto)
-			}
-			if err != nil {
-				pluginsState.returnCode = PluginsReturnCodeParseError
-				pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-				return response
-			}
-			serverInfo.noticeBegin(proxy)
-			if serverProto == "udp" {
-				response, err = proxy.exchangeWithUDPServer(serverInfo, sharedKey, encryptedQuery, clientNonce)
-				retryOverTCP := false
-				if err == nil && len(response) >= MinDNSPacketSize && response[2]&0x02 == 0x02 {
-					retryOverTCP = true
-				} else if neterr, ok := err.(net.Error); ok && neterr.Timeout() {
-					dlog.Debugf("[%v] Retry over TCP after UDP timeouts", serverName)
-					retryOverTCP = true
-				}
-				if retryOverTCP {
-					serverProto = "tcp"
-					sharedKey, encryptedQuery, clientNonce, err = proxy.Encrypt(serverInfo, query, serverProto)
-					if err != nil {
-						pluginsState.returnCode = PluginsReturnCodeParseError
-						pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-						return response
-					}
-					response, err = proxy.exchangeWithTCPServer(serverInfo, sharedKey, encryptedQuery, clientNonce)
-				}
-			} else {
-				response, err = proxy.exchangeWithTCPServer(serverInfo, sharedKey, encryptedQuery, clientNonce)
-			}
-			if err != nil {
-				if stale, ok := pluginsState.sessionData["stale"]; ok {
-					dlog.Debug("Serving stale response")
-					response, err = (stale.(*dns.Msg)).Pack()
-				}
-			}
-			if err != nil {
-				if neterr, ok := err.(net.Error); ok && neterr.Timeout() {
-					pluginsState.returnCode = PluginsReturnCodeServerTimeout
-				} else {
-					pluginsState.returnCode = PluginsReturnCodeNetworkError
-				}
-				pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-				serverInfo.noticeFailure(proxy)
-				return response
-			}
-		} else if serverInfo.Proto == stamps.StampProtoTypeDoH {
-			tid := TransactionID(query)
-			SetTransactionID(query, 0)
-			serverInfo.noticeBegin(proxy)
-			serverResponse, _, tls, _, err := proxy.xTransport.DoHQuery(serverInfo.useGet, serverInfo.URL, query, proxy.timeout)
-			SetTransactionID(query, tid)
 
-			if err != nil || tls == nil || !tls.HandshakeComplete {
-				if stale, ok := pluginsState.sessionData["stale"]; ok {
-					dlog.Debug("Serving stale response")
-					response, err = (stale.(*dns.Msg)).Pack()
-				}
-			}
-			if err != nil {
-				pluginsState.returnCode = PluginsReturnCodeNetworkError
-				pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-				serverInfo.noticeFailure(proxy)
-				return response
-			}
-			if response == nil {
-				response = serverResponse
-			}
-			if len(response) >= MinDNSPacketSize {
-				SetTransactionID(response, tid)
-			}
-		} else if serverInfo.Proto == stamps.StampProtoTypeODoHTarget {
-			tid := TransactionID(query)
-			if len(serverInfo.odohTargetConfigs) == 0 {
-				return response
-			}
-			target := serverInfo.odohTargetConfigs[rand.Intn(len(serverInfo.odohTargetConfigs))]
-			odohQuery, err := target.encryptQuery(query)
-			if err != nil {
-				dlog.Errorf("Failed to encrypt query for [%v]", serverName)
-				response = nil
-			} else {
-				targetURL := serverInfo.URL
-				if serverInfo.Relay != nil && serverInfo.Relay.ODoH != nil {
-					targetURL = serverInfo.Relay.ODoH.URL
-				}
-				responseBody, responseCode, _, _, err := proxy.xTransport.ObliviousDoHQuery(serverInfo.useGet, targetURL, odohQuery.odohMessage, proxy.timeout)
-				if err == nil && len(responseBody) > 0 && responseCode == 200 {
-					response, err = odohQuery.decryptResponse(responseBody)
-					if err != nil {
-						dlog.Warnf("Failed to decrypt response from [%v]", serverName)
-						response = nil
-					}
-				} else if responseCode == 401 || (responseCode == 200 && len(responseBody) == 0) {
-					if responseCode == 200 {
-						dlog.Warnf("ODoH relay for [%v] is buggy and returns a 200 status code instead of 401 after a key update", serverInfo.Name)
-					}
-					dlog.Infof("Forcing key update for [%v]", serverInfo.Name)
-					for _, registeredServer := range proxy.serversInfo.registeredServers {
-						if registeredServer.name == serverInfo.Name {
-							if err = proxy.serversInfo.refreshServer(proxy, registeredServer.name, registeredServer.stamp); err != nil {
-								// Failed to refresh the proxy server information.
-								dlog.Noticef("Key update failed for [%v]", serverName)
-								serverInfo.noticeFailure(proxy)
-								clocksmith.Sleep(10 * time.Second)
-							}
-							break
-						}
-					}
-					response = nil
-				} else {
-					dlog.Warnf("Failed to receive successful response from [%v]", serverName)
-				}
-			}
-
-			if len(response) >= MinDNSPacketSize {
-				SetTransactionID(response, tid)
-			} else if response == nil {
-				pluginsState.returnCode = PluginsReturnCodeNetworkError
-				pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-				serverInfo.noticeFailure(proxy)
-				return response
-			}
-		} else {
-			dlog.Fatal("Unsupported protocol")
-		}
-		if len(response) < MinDNSPacketSize || len(response) > MaxDNSPacketSize {
-			pluginsState.returnCode = PluginsReturnCodeParseError
-			pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-			serverInfo.noticeFailure(proxy)
+		// Exchange DNS request with the server
+		exchangeResponse, err := handleDNSExchange(proxy, serverInfo, &pluginsState, query, serverProto)
+		if err != nil || exchangeResponse == nil {
 			return response
 		}
-		response, err = pluginsState.ApplyResponsePlugins(&proxy.pluginsGlobals, response, ttl)
+
+		response = exchangeResponse
+
+		// Process the response through plugins
+		processedResponse, err := processPlugins(proxy, &pluginsState, query, serverInfo, response)
 		if err != nil {
-			pluginsState.returnCode = PluginsReturnCodeParseError
-			pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-			serverInfo.noticeFailure(proxy)
 			return response
 		}
-		if pluginsState.action == PluginsActionDrop {
-			pluginsState.returnCode = PluginsReturnCodeDrop
-			pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-			return response
-		}
-		if pluginsState.synthResponse != nil {
-			response, err = pluginsState.synthResponse.PackBuffer(response)
-			if err != nil {
-				pluginsState.returnCode = PluginsReturnCodeParseError
-				pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-				return response
-			}
-		}
-		if rcode := Rcode(response); rcode == dns.RcodeServerFailure { // SERVFAIL
-			if pluginsState.dnssec {
-				dlog.Debug("A response had an invalid DNSSEC signature")
-			} else {
-				dlog.Infof("A response with status code 2 was received - this is usually a temporary, remote issue with the configuration of the domain name")
-				serverInfo.noticeFailure(proxy)
-			}
-		} else {
-			serverInfo.noticeSuccess(proxy)
-		}
+
+		response = processedResponse
 	}
+
+	// Validate the response before sending
 	if len(response) < MinDNSPacketSize || len(response) > MaxDNSPacketSize {
 		if len(response) == 0 {
 			pluginsState.returnCode = PluginsReturnCodeNotReady
@@ -836,37 +759,16 @@ func (proxy *Proxy) processIncomingQuery(
 		}
 		return response
 	}
-	if clientProto == "udp" {
-		if len(response) > pluginsState.maxUnencryptedUDPSafePayloadSize {
-			response, err = TruncatedResponse(response)
-			if err != nil {
-				pluginsState.returnCode = PluginsReturnCodeParseError
-				pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-				return response
-			}
-		}
-		clientPc.(net.PacketConn).WriteTo(response, *clientAddr)
-		if HasTCFlag(response) {
-			proxy.questionSizeEstimator.blindAdjust()
-		} else {
-			proxy.questionSizeEstimator.adjust(ResponseOverhead + len(response))
-		}
-	} else if clientProto == "tcp" {
-		response, err = PrefixWithSize(response)
-		if err != nil {
-			pluginsState.returnCode = PluginsReturnCodeParseError
-			pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
-			if serverInfo != nil {
-				serverInfo.noticeFailure(proxy)
-			}
-			return response
-		}
-		if clientPc != nil {
-			clientPc.Write(response)
-		}
-	}
+
+	// Send the response back to the client
+	sendResponse(proxy, &pluginsState, response, clientProto, clientAddr, clientPc)
+
+	// Apply logging plugins
 	pluginsState.ApplyLoggingPlugins(&proxy.pluginsGlobals)
 
+	// Update monitoring metrics
+	updateMonitoringMetrics(proxy, &pluginsState, start)
+
 	return response
 }
 
